"""
audfprint2.py

Implementation of acoustic-landmark-based robust fingerprinting.
Port of the Matlab implementation.

2014-05-25 Dan Ellis dpwe@ee.columbia.edu
"""

import argparse
import concurrent.futures
import contextlib
import logging
import multiprocessing
import os
import time
from collections.abc import Iterable, Iterator, Sequence
from itertools import repeat
from multiprocessing.connection import Connection
from typing import Any, Callable, cast

import numpy as np

# My hash_table implementation
from audfprint2.core import analyzer, hash_table, matcher

time_clock = time.process_time
logger = logging.getLogger("audfprint2")


def filename_list_iterator(
    filelist: Sequence[str],
    wavdir: str,
    wavext: str,
    listflag: bool,
) -> Iterator[str]:
    """ Iterator to yeild all the filenames, possibly interpreting them
        as list files, prepending wavdir """
    if not listflag:
        for filename in filelist:
            yield os.path.join(wavdir, filename + wavext)
    else:
        for listfilename in filelist:
            with open(listfilename, 'r') as f:
                for filename in f:
                    yield os.path.join(wavdir, filename.rstrip('\n') + wavext)


# for saving precomputed fprints
def ensure_dir(dirname: str) -> None:
    """ ensure that the named directory exists """
    if len(dirname) and not os.path.exists(dirname):
        with contextlib.suppress(OSError):
            os.makedirs(dirname)


# Command line interface

# basic operations, each in a separate function

def file_precompute_peaks_or_hashes(
    analyzer_obj: analyzer.Analyzer,
    filename: str,
    precompdir: str,
    precompext: str | None = None,
    hashes_not_peaks: bool = True,
    skip_existing: bool = False,
    strip_prefix: str | None = None,
) -> list[str]:
    """ Perform precompute action for one file, return list
        of message strings """
    # If strip_prefix is specified and matches the start of filename,
    # remove it from filename.
    if strip_prefix and filename[:len(strip_prefix)] == strip_prefix:
        tail_filename = filename[len(strip_prefix):]
    else:
        tail_filename = filename
    # Form the output filename to check if it exists.
    # strip relative directory components from file name
    # Also remove leading absolute path (comp == '')
    relname = '/'.join(
        [
            comp
            for comp in tail_filename.split('/')
            if comp not in ['.', '..', '']
        ]
    )
    root = os.path.splitext(relname)[0]
    if precompext is None:
        precompext = analyzer.PRECOMPEXT if hashes_not_peaks else analyzer.PRECOMPPKEXT
    opfname = os.path.join(precompdir, root + precompext)
    if skip_existing and os.path.isfile(opfname):
        return [f"file {opfname} exists (and --skip-existing); skipping"]
    # Do the analysis
    saver: Callable[[str, Iterable[tuple[int, int]]], None]
    output: np.ndarray[Any, Any] | list[tuple[int, int]]
    if hashes_not_peaks:
        feature_type = "hashes"
        saver = analyzer.hashes_save
        output = analyzer_obj.wavfile2hashes(filename)
    else:
        feature_type = "peaks"
        saver = analyzer.peaks_save
        peaks = analyzer_obj.wavfile2peaks(filename)
        if peaks and isinstance(peaks[0], list):
            peak_lists = cast(list[list[tuple[int, int]]], peaks)
            output = [pair for peak_list in peak_lists for pair in peak_list]
        else:
            output = cast(list[tuple[int, int]], peaks)
    # save the hashes or peaks file
    if len(output) == 0:
        message = f"Zero length analysis for {filename} -- not saving."
    else:
        # Make sure the output directory exists
        ensure_dir(os.path.split(opfname)[0])
        # Write the file
        save_data: Iterable[tuple[int, int]]
        if isinstance(output, np.ndarray):
            save_data = [(int(pair[0]), int(pair[1])) for pair in output.tolist()]
        else:
            save_data = output
        saver(opfname, save_data)
        message = f"wrote {opfname}" + " ( %d %s, %.3f sec)" % (
            len(output),
            feature_type,
            analyzer_obj.soundfiledur,
        )
    return [message]


def file_precompute(
    analyzer_obj: analyzer.Analyzer,
    filename: str,
    precompdir: str,
    type: str = 'peaks',
    skip_existing: bool = False,
    strip_prefix: str | None = None,
) -> list[str]:
    """ Perform precompute action for one file, return list
        of message strings """
    logger.debug(time.ctime(), "precomputing", type, "for", filename, "...")
    hashes_not_peaks = (type == 'hashes')
    return file_precompute_peaks_or_hashes(analyzer_obj, filename, precompdir,
                                           hashes_not_peaks=hashes_not_peaks,
                                           skip_existing=skip_existing,
                                           strip_prefix=strip_prefix)


def make_ht_from_list(
    analyzer_obj: analyzer.Analyzer,
    filelist: Sequence[str],
    hashbits: int,
    depth: int,
    maxtime: int,
    pipe: Connection | None = None,
) -> hash_table.HashTable | None:
    """ Populate a hash table from a list, used as target for
        multiprocess division.  pipe is a pipe over which to push back
        the result, else return it """
    # Create new ht instance
    ht = hash_table.HashTable(hashbits=hashbits, depth=depth, maxtime=maxtime)
    # Add in the files
    for filename in filelist:
        hashes = analyzer_obj.wavfile2hashes(filename)
        hash_pairs = [
            (int(pair[0]), int(pair[1])) for pair in hashes.tolist()
        ] if isinstance(hashes, np.ndarray) else list(hashes)
        ht.store(filename, hash_pairs)
    # Pass back to caller
    if pipe:
        pipe.send(ht)
    else:
        return ht
    return None


def do_cmd(
    cmd: str,
    analyzer_obj: analyzer.Analyzer | None,
    hash_tab: hash_table.HashTable | None,
    filename_iter: Iterable[str],
    matcher_obj: matcher.Matcher | None,
    outdir: str,
    type: str,
    skip_existing: bool = False,
    strip_prefix: str | None = None,
) -> None:
    """ Breaks out the core part of running the command.
        This is just the single-core versions.
    """
    if cmd in {'merge', 'newmerge'}:
        if hash_tab is None:
            raise ValueError("hash table required for merge")
        # files are other hash tables, merge them in
        for filename in filename_iter:
            hash_tab2 = hash_table.HashTable(filename)
            if "samplerate" in hash_tab.params:
                assert hash_tab.params["samplerate"] == hash_tab2.params["samplerate"]
            else:
                # "newmerge" fails to setup the samplerate param
                hash_tab.params["samplerate"] = hash_tab2.params["samplerate"]
            hash_tab.merge(hash_tab2)

    elif cmd == 'precompute':
        if analyzer_obj is None:
            raise ValueError("analyzer required for precompute")
        # just precompute fingerprints, single core
        for filename in filename_iter:
            logger.debug(
                file_precompute(
                    analyzer_obj, filename, outdir, type,
                    skip_existing=skip_existing, strip_prefix=strip_prefix
                )
            )

    elif cmd == 'match':
        if analyzer_obj is None or matcher_obj is None or hash_tab is None:
            raise ValueError("analyzer, matcher, and hash table required for match")
        # Running query, single-core mode
        for num, filename in enumerate(filename_iter):
            msgs = matcher_obj.file_match_to_msgs(analyzer_obj, hash_tab, filename, num)
            for msg in msgs:
                logger.debug(msg)

    elif cmd in {'new', 'add'}:
        if analyzer_obj is None or hash_tab is None:
            raise ValueError("analyzer and hash table required for add/new")
        # Adding files
        tothashes = 0
        for ix, filename in enumerate(filename_iter):
            logger.debug(
                f"{time.ctime()} ingesting #{ix}: {filename} ..."
            )
            dur, nhash = analyzer_obj.ingest(hash_tab, filename)
            tothashes += nhash

        logger.debug(
            f"Added {tothashes} hashes "
            f"({tothashes / float(analyzer_obj.soundfiletotaldur):.1f} hashes/sec)"
        )
    elif cmd == 'remove':
        if hash_tab is None:
            raise ValueError("hash table required for remove")
        # Removing files from hash table.
        for filename in filename_iter:
            hash_tab.remove(filename)

    elif cmd == 'list':
        if hash_tab is None:
            raise ValueError("hash table required for list")
        hash_tab.list(lambda x: logger.debug([x]))

    else:
        raise ValueError(f"unrecognized command: {cmd}")


def multiproc_add(
    analyzer_obj: analyzer.Analyzer,
    hash_tab: hash_table.HashTable,
    filename_iter: Iterable[str],
    report: Any | None,
    ncores: int,
) -> None:
    """Run multiple threads adding new files to hash table"""
    # run ncores in parallel to add new files to existing HASH_TABLE
    # lists store per-process parameters
    # Pipes to transfer results
    rx: list[Connection[Any, Any]] = []
    tx: list[Connection[Any, Any]] = []
    # Process objects
    pr: list[multiprocessing.Process] = []
    # Lists of the distinct files
    filelists: list[list[str]] = [[] for _ in range(ncores)]
    # unpack all the files into ncores lists
    for ix, filename in enumerate(filename_iter):
        filelists[ix % ncores].append(filename)
    # Launch each of the individual processes
    for ix in range(ncores):
        recv_conn, send_conn = multiprocessing.Pipe(False)
        rx.append(recv_conn)
        tx.append(send_conn)
        process = multiprocessing.Process(target=make_ht_from_list,
                                          args=(analyzer_obj, filelists[ix],
                                                hash_tab.hashbits,
                                                hash_tab.depth,
                                                (1 << hash_tab.maxtimebits),
                                                send_conn))
        pr.append(process)
        process.start()
    # gather results when they all finish
    for core, recv_conn in enumerate(rx):
        # thread passes back serialized hash table structure
        hash_tabx = recv_conn.recv()
        logger.debug(
            f"hash_table {core} has {len(hash_tabx.names)} files "
            f"{sum(hash_tabx.counts)} hashes"
        )
        # merge in all the new items, hash entries
        hash_tab.merge(hash_tabx)
        # finish that thread...
        pr[core].join()


def matcher_file_match_to_msgs(
    matcher_obj: matcher.Matcher,
    analyzer_obj: analyzer.Analyzer,
    hash_tab: hash_table.HashTable,
    filename: str,
) -> list[str]:
    """Cover for matcher.file_match_to_msgs so it can be passed to a process pool"""
    return matcher_obj.file_match_to_msgs(analyzer_obj, hash_tab, filename)


def do_cmd_multiproc(
    cmd: str,
    analyzer_obj: analyzer.Analyzer | None,
    hash_tab: hash_table.HashTable | None,
    filename_iter: Iterable[str],
    matcher_obj: matcher.Matcher | None,
    outdir: str,
    type: str,
    report: Any | None = None,
    skip_existing: bool = False,
    strip_prefix: str | None = None,
    ncores: int = 1,
) -> None:
    """ Run the actual command, using multiple processors """
    if cmd == 'precompute':
        if analyzer_obj is None:
            raise ValueError("analyzer required for multiprocess operations")
        # precompute fingerprints with concurrent futures
        with concurrent.futures.ProcessPoolExecutor(max_workers=ncores) as executor:
            msgs_iter = executor.map(
                file_precompute,
                repeat(analyzer_obj),
                filename_iter,
                repeat(outdir),
                repeat(type),
                repeat(skip_existing),
                repeat(strip_prefix),
            )
            for msgs in msgs_iter:
                logger.debug(msgs)

    elif cmd in {'new', 'add'}:
        # Ingest files into the hash table in parallel
        if analyzer_obj is None or hash_tab is None:
            raise ValueError("analyzer and hash table required for add/new")
        multiproc_add(analyzer_obj, hash_tab, filename_iter, report, ncores)

    elif cmd == 'match':
        if analyzer_obj is None or matcher_obj is None or hash_tab is None:
            raise ValueError("analyzer, matcher, and hash table required for match")
        # Running queries in parallel
        with concurrent.futures.ProcessPoolExecutor(max_workers=ncores) as executor:
            msgs_iter = executor.map(
                matcher_file_match_to_msgs,
                repeat(matcher_obj),
                repeat(analyzer_obj),
                repeat(hash_tab),
                filename_iter,
            )
            for msgs in msgs_iter:
                logger.debug(msgs)

    else:
        # This is not a multiproc command
        raise ValueError(f"unrecognized multiproc command: {cmd}")


# Command to separate out setting of analyzer parameters
def setup_analyzer(
    density: float,
    is_match: bool,
    pks_per_frame: int,
    fanout: int,
    freq_sd: float,
    shifts: int,
    samplerate: int,
    continue_on_error: bool,
    ffmpeg_path: str | None,
) -> analyzer.Analyzer:
    """Create a new analyzer object, taking values from docopts args"""
    # Create analyzer object; parameters will get set below
    analyzer_obj = analyzer.Analyzer()
    # Read parameters from command line/docopts
    analyzer_obj.density = density
    analyzer_obj.maxpksperframe = pks_per_frame
    analyzer_obj.maxpairsperpeak = fanout
    analyzer_obj.f_sd = freq_sd
    analyzer_obj.shifts = shifts
    # fixed - 512 pt FFT with 256 pt hop at 11025 Hz
    analyzer_obj.target_sr = samplerate
    analyzer_obj.n_fft = 512
    analyzer_obj.n_hop = analyzer_obj.n_fft // 2
    # set default value for shifts depending on mode
    if analyzer_obj.shifts == 0:
        # Default shift is 4 for match, otherwise 1
        analyzer_obj.shifts = 4 if is_match else 1
    analyzer_obj.fail_on_error = not continue_on_error
    analyzer_obj.ffmpeg_path = ffmpeg_path
    return analyzer_obj


def setup_matcher(
    match_win: int,
    search_depth: int,
    min_count: int,
    max_matches: int,
    exact_count: bool,
    find_time_range: bool,
    time_quantile: float,
    sortbytime: bool,
    illustrate: bool,
    illustrate_hpf: bool,
) -> matcher.Matcher:
    """Create a new matcher objects, set parameters from docopt structure"""
    matcher_obj = matcher.Matcher()
    matcher_obj.window = match_win
    matcher_obj.threshcount = min_count
    matcher_obj.max_returns = max_matches
    matcher_obj.search_depth = search_depth
    matcher_obj.sort_by_time = sortbytime
    matcher_obj.exact_count = exact_count | illustrate | illustrate_hpf
    matcher_obj.illustrate = illustrate | illustrate_hpf
    matcher_obj.illustrate_hpf = illustrate_hpf
    matcher_obj.find_time_range = find_time_range
    matcher_obj.time_quantile = time_quantile
    return matcher_obj


__version__ = 20251119


def build_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        description=(
            "Landmark-based audio fingerprinting. Create a new fingerprint dbase with 'new', append "
            "new files to an existing database with 'add', or identify noisy query excerpts with "
            "'match'. 'precompute' writes a *.fpt file under precompdir with precomputed fingerprint "
            "for each input wav file. 'merge' combines previously-created databases into an existing "
            "database; 'newmerge' combines existing databases to create a new one."
        ),
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        add_help=False,
    )
    parser.add_argument("-?", "--help", action="help", help="Show this help message and exit.")
    parser.add_argument(
        "--version",
        action="version",
        version=f"audfprint2-enhanced {__version__}",
        help="Show the program's version number and exit.",
    )
    parser.add_argument(
        "cmd",
        choices=["new", "add", "precompute", "merge", "newmerge", "match", "list", "remove"],
        help="Command to execute.",
    )
    parser.add_argument("-d", "--dbase", help="Fingerprint database file")
    parser.add_argument("-n", "--density", type=float, default=20.0, help="Target hashes per second")
    parser.add_argument("-h", "--hashbits", type=int, default=20, help="How many bits in each hash")
    parser.add_argument("-b", "--bucketsize", type=int, default=100, help="Number of entries per bucket")
    parser.add_argument("-t", "--maxtime", type=int, default=16384, help="Largest time value stored")
    parser.add_argument(
        "-u",
        "--maxtimebits",
        type=int,
        help="maxtime as a number of bits (16384 == 14 bits)",
    )
    parser.add_argument(
        "-r", "--samplerate", type=int, default=11025,
        help="Resample input files to this"
    )
    parser.add_argument(
        "--ffmpeg-path",
        default=None,
        help="Explicit path to ffmpeg binary (overrides PATH)",
    )
    parser.add_argument(
        "-p", "--precompdir", default=".",
        help="Save precomputed files under this dir"
    )
    parser.add_argument(
        "-i", "--shifts", type=int, default=0,
        help="Use this many subframe shifts building fp"
    )
    parser.add_argument(
        "-w", "--match-win", type=int, default=2,
        help="Maximum tolerable frame skew to count as a match"
    )
    parser.add_argument(
        "-N", "--min-count", type=int, default=5,
        help="Minimum number of matching landmarks to count as a match"
    )
    parser.add_argument(
        "-x", "--max-matches", type=int, default=1,
        help="Maximum number of matches to report for each query"
    )
    parser.add_argument(
        "-X", "--exact-count", action="store_true",
        help="Flag to use more precise (but slower) match counting"
    )
    parser.add_argument(
        "-R", "--find-time-range", action="store_true",
        help="Report the time support of each match"
    )
    parser.add_argument(
        "-Q", "--time-quantile", type=float, default=0.05,
        help="Quantile at extremes of time support"
    )
    parser.add_argument(
        "-S", "--freq-sd", type=float, default=30.0,
        help="Frequency peak spreading SD in bins"
    )
    parser.add_argument(
        "-F", "--fanout", type=int, default=3,
        help="Max number of hash pairs per peak"
    )
    parser.add_argument(
        "-P", "--pks-per-frame", type=int, default=5,
        help="Maximum number of peaks per frame"
    )
    parser.add_argument(
        "-D", "--search-depth", type=int, default=100,
        help="How far down to search raw matching track list"
    )
    parser.add_argument(
        "-H", "--ncores", type=int, default=1,
        help="Number of processes to use"
    )
    parser.add_argument(
        "-o", "--opfile", default="",
        help="Write output (matches) to this file, not stdout"
    )
    parser.add_argument(
        "-K", "--precompute-peaks", action="store_true",
        help="Precompute just landmarks (else full hashes)"
    )
    parser.add_argument(
        "-k", "--skip-existing", action="store_true",
        help="On precompute, skip items if output file already exists"
    )
    parser.add_argument(
        "-C", "--continue-on-error", action="store_true",
        help="Keep processing despite errors reading input"
    )
    parser.add_argument(
        "-l", "--list", dest="list", action="store_true",
        help="Input files are lists, not audio"
    )
    parser.add_argument(
        "-T", "--sortbytime", action="store_true",
        help="Sort multiple hits per file by time (instead of score)"
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="Set logging level to debug (default is warning)",
    )
    parser.add_argument(
        "-I", "--illustrate", action="store_true",
        help="Make a plot showing the match"
    )
    parser.add_argument(
        "-J", "--illustrate-hpf", action="store_true",
        help="Plot the match, using onset enhancement"
    )
    parser.add_argument(
        "-W", "--wavdir", default="",
        help="Find sound files under this dir"
    )
    parser.add_argument(
        "-V", "--wavext", default="",
        help="Extension to add to wav file names"
    )
    parser.add_argument("file", nargs="*", help="Files or lists (with --list) to process")
    return parser


def parse_args(argv: Sequence[str] | None = None) -> argparse.Namespace:
    """Create and parse the command-line arguments."""
    parser = build_parser()
    return parser.parse_args(argv)


def main(argv: Sequence[str] | None = None) -> None:
    args = parse_args(argv)
    logging.getLogger().setLevel(logging.DEBUG if args.verbose else logging.WARNING)

    initticks = time_clock()

    maxtimebits = args.maxtimebits if args.maxtimebits is not None else hash_table._bitsfor(args.maxtime)

    # Setup the analyzer if we're using one (i.e., unless "merge")
    analyzer_obj = setup_analyzer(
        args.density,
        args.cmd == "match",
        args.pks_per_frame,
        args.fanout,
        args.freq_sd,
        args.shifts,
        args.samplerate,
        args.continue_on_error,
        args.ffmpeg_path,
    ) if args.cmd not in ["merge", "newmerge", "list", "remove"] else None

    precomp_type = 'peaks' if args.precompute_peaks else 'hashes'

    # Set up the hash table, if we're using one (i.e., unless "precompute")
    if args.cmd != "precompute":
        # For everything other than precompute, we need a database name
        # Check we have one
        if not args.dbase:
            raise ValueError("dbase name must be provided if not precompute")
        if args.cmd in ["new", "newmerge"]:
            # Check that the output directory can be created before we start
            ensure_dir(os.path.split(args.dbase)[0])
            # Create a new hash table
            hash_tab = hash_table.HashTable(
                    hashbits=args.hashbits,
                    depth=args.bucketsize,
                    maxtime=(1 << maxtimebits))
            # Set its samplerate param
            if analyzer_obj:
                hash_tab.params['samplerate'] = analyzer_obj.target_sr

        else:
            logger.debug(f"{time.ctime()} Reading hash table {args.dbase}")
            hash_tab = hash_table.HashTable(args.dbase)
            if analyzer_obj and 'samplerate' in hash_tab.params \
                    and hash_tab.params['samplerate'] != analyzer_obj.target_sr:
                logger.debug("db samplerate overridden to ", analyzer_obj.target_sr)
    else:
        # The command IS precompute
        # dummy empty hash table
        hash_tab = None

    # Create a matcher
    matcher_obj = setup_matcher(
        args.match_win,
        args.search_depth,
        args.min_count,
        args.max_matches,
        args.exact_count,
        args.find_time_range,
        args.time_quantile,
        args.sortbytime,
        args.illustrate,
        args.illustrate_hpf,
    ) if args.cmd == 'match' else None

    filename_iter = filename_list_iterator(
            args.file, args.wavdir, args.wavext, args.list)

    #######################
    # Run the main commmand
    #######################

    # How many processors to use (multiprocessing)
    if args.ncores > 1 and args.cmd not in ["merge", "newmerge", "list", "remove"]:
        # merge/newmerge/list/remove are always single-thread processes
        do_cmd_multiproc(args.cmd, analyzer_obj, hash_tab, filename_iter,
                         matcher_obj, args.precompdir,
                         precomp_type,
                         skip_existing=args.skip_existing,
                         strip_prefix=args.wavdir,
                         ncores=args.ncores)
    else:
        do_cmd(args.cmd, analyzer_obj, hash_tab, filename_iter,
               matcher_obj, args.precompdir, precomp_type,
               skip_existing=args.skip_existing,
               strip_prefix=args.wavdir)

    elapsedtime = time_clock() - initticks
    if analyzer_obj and analyzer_obj.soundfiletotaldur > 0.:
        logger.debug("Processed "
              + "%d files (%.1f s total dur) in %.1f s sec = %.3f x RT" \
              % (analyzer_obj.soundfilecount, analyzer_obj.soundfiletotaldur,
                 elapsedtime, (elapsedtime / analyzer_obj.soundfiletotaldur)))

    # Save the hash table file if it has been modified
    if hash_tab and hash_tab.dirty:
        # We already created the directory, if "new".
        hash_tab.save(args.dbase)


if __name__ == '__main__':
    main()



"""Provide stft to avoid librosa dependency.

This implementation is based on routines from
https://github.com/tensorflow/models/blob/master/research/audioset/mel_features.py
"""

from typing import Literal

import numpy as np


def frame(data: np.ndarray, window_length: int, hop_length: int) -> np.ndarray:
  """Convert array into a sequence of successive possibly overlapping frames.

  An n-dimensional array of shape (num_samples, ...) is converted into an
  (n+1)-D array of shape (num_frames, window_length, ...), where each frame
  starts hop_length points after the preceding one.

  This is accomplished using stride_tricks, so the original data is not
  copied.  However, there is no zero-padding, so any incomplete frames at the
  end are not included.

  Args:
    data: np.array of dimension N >= 1.
    window_length: Number of samples in each frame.
    hop_length: Advance (in samples) between each window.

  Returns:
    (N+1)-D np.array with as many rows as there are complete frames that can be
    extracted.
  """
  num_samples = data.shape[0]
  num_frames = 1 + ((num_samples - window_length) // hop_length)
  shape = (num_frames, window_length) + data.shape[1:]
  strides = (data.strides[0] * hop_length,) + data.strides
  return np.lib.stride_tricks.as_strided(data, shape=shape, strides=strides)


def periodic_hann(window_length: int) -> np.ndarray:
  """Calculate a "periodic" Hann window.

  The classic Hann window is defined as a raised cosine that starts and
  ends on zero, and where every value appears twice, except the middle
  point for an odd-length window.  Matlab calls this a "symmetric" window
  and np.hanning() returns it.  However, for Fourier analysis, this
  actually represents just over one cycle of a period N-1 cosine, and
  thus is not compactly expressed on a length-N Fourier basis.  Instead,
  it's better to use a raised cosine that ends just before the final
  zero value - i.e. a complete cycle of a period-N cosine.  Matlab
  calls this a "periodic" window. This routine calculates it.

  Args:
    window_length: The number of points in the returned window.

  Returns:
    A 1D np.array containing the periodic hann window.
  """
  return 0.5 - (0.5 * np.cos(2 * np.pi / window_length *
                             np.arange(window_length)))


def stft(
    signal: np.ndarray,
    n_fft: int,
    hop_length: int | None = None,
    window: int | float | np.ndarray | None = None,
) -> np.ndarray:
  """Calculate the short-time Fourier transform.

  Args:
    signal: 1D np.array of the input time-domain signal.
    n_fft: Size of the FFT to apply.
    hop_length: Advance (in samples) between each frame passed to FFT. Defaults
      to half the window length.
    window: Length of each block of samples to pass to FFT, or vector of window
      values.  Defaults to n_fft.

  Returns:
    2D np.array where each column contains the complex values of the
    fft_length/2+1 unique values of the FFT for the corresponding frame of
    input samples ("spectrogram transposition").
  """
  if window is None:
    window = n_fft
  if isinstance(window, (int, float)):
    # window holds the window length, need to make the actual window.
    window = periodic_hann(int(window))
  window_length = len(window)
  if not hop_length:
    hop_length = window_length // 2
  # Default librosa STFT behavior.
  pad_mode: Literal['reflect'] = 'reflect'
  signal = np.pad(signal, (n_fft // 2), mode=pad_mode)
  frames = frame(signal, window_length, hop_length)
  # Apply frame window to each frame. We use a periodic Hann (cosine of period
  # window_length) instead of the symmetric Hann of np.hanning (period
  # window_length-1).
  windowed_frames = frames * window
  return np.fft.rfft(windowed_frames, n_fft).transpose()



# coding=utf-8
"""audio_read reads in a whole audio file with resampling."""

# Equivalent to:
# import librosa
# def audio_read(filename, sr=11025, channels=1):
#    """Read a soundfile, return (d, sr)."""
#    d, sr = librosa.load(filename, sr=sr, mono=(channels == 1))
#    return d, sr

# The code below is adapted from:
# https://github.com/bmcfee/librosa/blob/master/librosa/core/audio.py
# This is its original copyright notice:

# Copyright (c) 2014, Brian McFee, Matt McVicar, Dawen Liang, Colin Raffel, Douglas Repetto, Dan Ellis.
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
# AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
# INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
# LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
# OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
# PERFORMANCE OF THIS SOFTWARE.

import os
import queue
import re
import subprocess
import threading
import time
from collections.abc import Iterator
from types import TracebackType
from typing import IO, Literal

import numpy as np

# For wavread fallback.
import scipy.io.wavfile as wav  # type: ignore[import-untyped]

# If ffmpeg is unavailable, you can set HAVE_FFMPEG to False which will cause
# soundfile reads to go via scipy.io.wavfile.  However, this means that only
# *.wav files are supported *and* they must already be resampled to the
# system sampling rate (e.g. 11025 Hz).

HAVE_FFMPEG = True
DEFAULT_FFMPEG_BINARY = "ffmpeg"
# Allow overriding the ffmpeg binary via environment variable so users can provide
# a custom installation without modifying PATH.
FFMPEG_BINARY = os.environ.get("AUDFPRINT_FFMPEG", DEFAULT_FFMPEG_BINARY)


def set_ffmpeg_path(ffmpeg_path: str) -> None:
    """Set the global ffmpeg binary path used by audio readers."""
    global FFMPEG_BINARY
    FFMPEG_BINARY = ffmpeg_path


def wavread(filename: str) -> tuple[np.ndarray, int]:
    """Read in audio data from a wav file.  Return d, sr."""
    # Read in wav file.
    samplerate, wave_data = wav.read(filename)
    # Normalize short ints to floats in range [-1..1).
    data = np.asarray(wave_data, dtype=np.float64) / 32768.0
    return data, samplerate


def audio_read(
    filename: str,
    sr: int | None = None,
    channels: int | None = None,
    ffmpeg_path: str | None = None,
) -> tuple[np.ndarray, int]:
    """Read a soundfile, return (d, sr)."""
    if HAVE_FFMPEG:
        return audio_read_ffmpeg(filename, sr, channels, ffmpeg_path=ffmpeg_path)
    data, samplerate = wavread(filename)
    if channels == 1 and len(data.shape) == 2 and data.shape[-1] != 1:
        # Convert stereo to mono.
        data = np.mean(data, axis=-1)
    if sr and sr != samplerate:
        raise ValueError("Wav file has samplerate %f but %f requested." % (samplerate, sr))
    return data, samplerate


def audio_read_ffmpeg(
    filename: str,
    sr: int | None = None,
    channels: int | None = None,
    ffmpeg_path: str | None = None,
) -> tuple[np.ndarray, int]:
    """Read a soundfile, return (d, sr)."""
    # Hacked version of librosa.load and audioread/ff.
    offset = 0.0
    duration = None
    dtype = np.float32
    frames: list[np.ndarray] = []
    with FFmpegAudioFile(
        os.path.realpath(filename),
        sample_rate=sr,
        channels=channels,
        ffmpeg_path=ffmpeg_path,
    ) as input_file:
        sr = input_file.sample_rate
        channels = input_file.channels
        if sr is None or channels is None:
            raise ValueError("Sample rate or channels could not be determined.")
        s_start = int(np.floor(sr * offset) * channels)
        if duration is None:
            s_end = np.inf
        else:
            s_end = s_start + int(np.ceil(sr * duration) * channels)
        num_read = 0
        for frame_bytes in input_file:
            frame = buf_to_float(frame_bytes, dtype=dtype)
            num_read_prev = num_read
            num_read += len(frame)
            if num_read < s_start:
                # offset is after the current frame, keep reading.
                continue
            if s_end < num_read_prev:
                # we're off the end.  stop reading
                break
            if s_end < num_read:
                # the end is in this frame.  crop.
                frame = frame[: int(s_end - num_read_prev)]
            if num_read_prev <= s_start < num_read:
                # beginning is in this frame
                frame = frame[(s_start - num_read_prev) :]
            # tack on the current frame
            frames.append(frame)

        if not len(frames):
            # Zero-length read
            y: np.ndarray = np.zeros(0, dtype=dtype)
        else:
            y = np.concatenate(frames)
            if channels > 1:
                y = y.reshape((-1, 2)).T

    # Final cleanup for dtype and contiguity
    y = np.ascontiguousarray(y, dtype=dtype)

    return (y, sr)


def buf_to_float(
    x: bytes | bytearray | memoryview | np.ndarray,
    n_bytes: int = 2,
    dtype: np.dtype | type = np.float32,
) -> np.ndarray:
    """Convert an integer buffer to floating point values.
    This is primarily useful when loading integer-valued wav data
    into numpy arrays.
    .. seealso:: :func:`librosa.util.buf_to_float`
    :parameters:
        - x : np.ndarray [dtype=int]
            The integer-valued data buffer
        - n_bytes : int [1, 2, 4]
            The number of bytes per sample in ``x``
        - dtype : numeric type
            The target output type (default: 32-bit float)
    :return:
        - x_float : np.ndarray [dtype=float]
            The input data buffer cast to floating point
    """

    # Invert the scale of the data
    scale = 1.0 / float(1 << ((8 * n_bytes) - 1))

    # Construct the format string
    fmt = "<i{:d}".format(n_bytes)

    # Rescale and format the data buffer
    return scale * np.frombuffer(x, fmt).astype(dtype)


# The code below is adapted from:
# https://github.com/sampsyo/audioread/blob/master/audioread/ffdec.py
# Below is its original copyright notice:

# This file is part of audioread.
# Copyright 2014, Adrian Sampson.
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.


class QueueReaderThread(threading.Thread):
    """A thread that consumes data from a filehandle and sends the data
    over a Queue.
    """

    def __init__(self, fh: IO[bytes], blocksize: int = 1024, discard: bool = False) -> None:
        super(QueueReaderThread, self).__init__()
        self.fh = fh
        self.blocksize = blocksize
        self.daemon = True
        self.discard = discard
        self.queue: queue.Queue[bytes] = queue.Queue()

    def run(self) -> None:
        while True:
            data = self.fh.read(self.blocksize)
            if not self.discard:
                self.queue.put(data)
            if not data:
                # Stream closed (EOF).
                break


class FFmpegAudioFile(object):
    """An audio file decoded by the ffmpeg command-line utility."""

    def __init__(
        self,
        filename: str,
        channels: int | None = None,
        sample_rate: int | None = None,
        block_size: int = 4096,
        ffmpeg_path: str | None = None,
    ) -> None:
        if not os.path.isfile(filename):
            raise ValueError(f"{filename} not found.")
        ffmpeg_bin = ffmpeg_path or FFMPEG_BINARY
        popen_args = [ffmpeg_bin, "-i", filename, "-f", "s16le"]
        self.channels = channels
        self.sample_rate = sample_rate
        if channels:
            popen_args.extend(["-ac", str(channels)])
        if sample_rate:
            popen_args.extend(["-ar", str(sample_rate)])
        popen_args.append("-")
        self.proc = subprocess.Popen(popen_args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

        # Start another thread to consume the standard output of the
        # process, which contains raw audio data.
        if self.proc.stdout is None:
            raise ValueError("stdout stream unavailable")
        self.stdout_reader = QueueReaderThread(self.proc.stdout, block_size)
        self.stdout_reader.start()

        # Read relevant information from stderr.
        try:
            self._get_info()
        except ValueError as err:
            raise ValueError(f"Error reading header info from {filename}") from err

        # Start a separate thread to read the rest of the data from
        # stderr. This (a) avoids filling up the OS buffer and (b)
        # collects the error output for diagnosis.
        if self.proc.stderr is None:
            raise ValueError("stderr stream unavailable")
        self.stderr_reader = QueueReaderThread(self.proc.stderr)
        self.stderr_reader.start()

    def read_data(self, timeout: float = 10.0) -> Iterator[bytes]:
        """Read blocks of raw PCM data from the file."""
        # Read from stdout in a separate thread and consume data from
        # the queue.
        start_time = time.time()
        while True:
            # Wait for data to be available or a timeout.
            data = None
            try:
                data = self.stdout_reader.queue.get(timeout=timeout)
                if data:
                    yield data
                else:
                    # End of file.
                    break
            except queue.Empty as err:
                # Queue read timed out.
                end_time = time.time()
                if not data:
                    if end_time - start_time >= timeout:
                        # Nothing interesting has happened for a while --
                        # FFmpeg is probably hanging.
                        raise ValueError(
                            f'ffmpeg output: {b"".join(self.stderr_reader.queue.queue).decode("utf8", "ignore")}'
                        ) from err
                    start_time = end_time
                    # Keep waiting.
                    continue

    def _get_info(self) -> None:
        """Reads the tool's output from its stderr stream, extracts the
        relevant information, and parses it.
        """
        if self.proc.stderr is None:
            raise ValueError("stderr stream unavailable")
        stderr = self.proc.stderr
        out_parts = []
        while True:
            line_bytes = stderr.readline()
            if not line_bytes:
                # EOF and data not found.
                raise ValueError("stream info not found")

            line = line_bytes.decode("utf8", "ignore")

            line = line.strip().lower()

            if "no such file" in line:
                raise IOError("file not found")
            elif "invalid data found" in line:
                raise ValueError()
            elif "duration:" in line:
                out_parts.append(line)
            elif "audio:" in line:
                out_parts.append(line)
                self._parse_info("".join(out_parts))
                break

    def _parse_info(self, s: str) -> None:
        """Given relevant data from the ffmpeg output, set audio
        parameter fields on this object.
        """
        if match := re.search(r"(\d+) hz", s):
            self.sample_rate_orig = int(match[1])
        else:
            self.sample_rate_orig = 0
        if self.sample_rate is None:
            self.sample_rate = self.sample_rate_orig

        if match := re.search(r"hz, ([^,]+),", s):
            mode = match[1]
            if mode == "stereo":
                self.channels_orig = 2
            elif match := re.match(r"(\d+) ", mode):
                self.channels_orig = int(match[1])
            else:
                self.channels_orig = 1
        else:
            self.channels_orig = 0
        if self.channels is None:
            self.channels = self.channels_orig

        if match := re.search(r"duration: (\d+):(\d+):(\d+).(\d)", s):
            durparts = list(map(int, match.groups()))
            duration = durparts[0] * 60 * 60 + durparts[1] * 60 + durparts[2] + float(durparts[3]) / 10
            self.duration = duration
        else:
            # No duration found.
            self.duration = 0

    def close(self) -> None:
        """Close the ffmpeg process used to perform the decoding."""
        # Kill the process if it is still running.
        if hasattr(self, "proc") and self.proc.returncode is None:
            self.proc.kill()
            self.proc.wait()

    def __del__(self) -> None:
        self.close()

    # Iteration.
    def __iter__(self) -> Iterator[bytes]:
        return self.read_data()

    # Context manager.
    def __enter__(self) -> "FFmpegAudioFile":
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> Literal[False]:
        self.close()
        return False



# coding=utf-8
"""Landmark-based audio fingerprinting library."""

from audfprint2.core.analyzer import Analyzer
from audfprint2.core.hash_table import DatabaseType, HashTable
from audfprint2.core.matcher import Matcher

__version__ = "20251119"

__all__ = [
    "Analyzer",
    "Matcher",
    "HashTable",
    "DatabaseType",
    "__version__",
]



# coding=utf-8
"""
audfprint2_analyze.py

Class to do the analysis of wave files into hash constellations.

2014-09-20 Dan Ellis dpwe@ee.columbia.edu
"""

import glob
import logging
import os

# For reading/writing hashes to file
import struct
import time
from collections.abc import Iterable, Sequence
from typing import Any, cast

import numpy as np
import scipy  # type: ignore[import-untyped]
import scipy.signal  # type: ignore[import-untyped]

# For utility, glob2hashtable
from audfprint2.core import hash_table
from audfprint2.utils import audio, stft

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger("audfprint2")

# ############### Globals ############### #
# Special extension indicating precomputed fingerprint
PRECOMPEXT = '.afpt'
# A different precomputed fingerprint is just the peaks
PRECOMPPKEXT = '.afpk'


def locmax(vec: np.ndarray, indices: bool = False) -> np.ndarray:
    """ Return a boolean vector of which points in vec are local maxima.
        End points are peaks if larger than single neighbors.
        if indices=True, return the indices of the True values instead
        of the boolean vector.
    """
    # vec[-1]-1 means last value can be a peak
    # nbr = np.greater_equal(np.r_[vec, vec[-1]-1], np.r_[vec[0], vec])
    # the np.r_ was killing us, so try an optimization...
    nbr = np.zeros(len(vec) + 1, dtype=bool)
    nbr[0] = True
    nbr[1:-1] = np.greater_equal(vec[1:], vec[:-1])
    maxmask = (nbr[:-1] & ~nbr[1:])
    return np.nonzero(maxmask)[0] if indices else maxmask


# Constants for Analyzer
# DENSITY controls the density of landmarks found (approx DENSITY per sec)
DENSITY = 20.0
# OVERSAMP > 1 tries to generate extra landmarks by decaying faster
OVERSAMP = 1
# 512 pt FFT @ 11025 Hz, 50% hop
# t_win = 0.0464
# t_hop = 0.0232
# Just specify n_fft
N_FFT = 512
N_HOP = 256
# spectrogram enhancement
HPF_POLE = 0.98

# Globals defining packing of landmarks into hashes
F1_BITS = 8
DF_BITS = 6
DT_BITS = 6
# derived constants
B1_MASK = (1 << F1_BITS) - 1
B1_SHIFT = DF_BITS + DT_BITS
DF_MASK = (1 << DF_BITS) - 1
DF_SHIFT = DT_BITS
DT_MASK = (1 << DT_BITS) - 1


def landmarks2hashes(
    landmarks: np.ndarray | Sequence[Sequence[int]]
) -> np.ndarray:
    """Convert a list of (time, bin1, bin2, dtime) landmarks
    into a list of (time, hash) pairs where the hash combines
    the three remaining values.
    """
    landmarks = np.array(landmarks)
    # Deal with special case of empty landmarks.
    if landmarks.shape[0] == 0:
      return np.zeros((0, 2), dtype=np.int32)
    hashes = np.zeros((landmarks.shape[0], 2), dtype=np.int32)
    hashes[:, 0] = landmarks[:, 0]
    hashes[:, 1] = (((landmarks[:, 1] & B1_MASK) << B1_SHIFT)
                    | (((landmarks[:, 2] - landmarks[:, 1]) & DF_MASK)
                       << DF_SHIFT)
                    | (landmarks[:, 3] & DT_MASK))
    return hashes


def hashes2landmarks(
    hashes: Sequence[tuple[int, int]]
) -> list[tuple[int, int, int, int]]:
    """Convert the mashed-up landmarks in hashes back into a list
    of (time, bin1, bin2, dtime) tuples.
    """
    landmarks = []
    for time_, hash_ in hashes:
        dtime = hash_ & DT_MASK
        bin1 = (hash_ >> B1_SHIFT) & B1_MASK
        dbin = (hash_ >> DF_SHIFT) & DF_MASK
        # Sign extend frequency difference
        if dbin >= (1 << (DF_BITS - 1)):
            dbin -= (1 << DF_BITS)
        landmarks.append((time_, bin1, bin1 + dbin, dtime))
    return landmarks


class Analyzer(object):
    """ A class to wrap up all the parameters associated with
        the analysis of soundfiles into fingerprints """
    # Parameters

    # optimization: cache pre-calculated Gaussian profile
    __sp_width: float | None = None
    __sp_len: int | None = None
    __sp_vals: np.ndarray = np.array([])

    def __init__(self, density: float = DENSITY) -> None:
        self.density = density
        self.target_sr = 11025
        self.n_fft = N_FFT
        self.n_hop = N_HOP
        self.shifts = 1
        # how wide to spreak peaks
        self.f_sd = 30.0
        # Maximum number of local maxima to keep per frame
        self.maxpksperframe = 5
        # Limit the num of pairs we'll make from each peak (Fanout)
        self.maxpairsperpeak = 3
        # Values controlling peaks2landmarks
        # +/- 31 bins in freq (LIMITED TO -32..31 IN LANDMARK2HASH)
        self.targetdf = 31
        # min time separation (traditionally 1, upped 2014-08-04)
        self.mindt = 2
        # max lookahead in time (LIMITED TO <64 IN LANDMARK2HASH)
        self.targetdt = 63
        # global stores duration of most recently-read soundfile
        self.soundfiledur = 0.0
        # .. and total amount of sound processed
        self.soundfiletotaldur = 0.0
        # .. and count of files
        self.soundfilecount = 0
        # Control behavior on file reading error
        self.fail_on_error = True
        # Optional override for ffmpeg binary location
        self.ffmpeg_path: str | None = None

    def spreadpeaksinvector(self, vector: np.ndarray, width: float = 4.0) -> np.ndarray:
        """ Create a blurred version of vector, where each of the local maxes
            is spread by a gaussian with SD <width>.
        """
        npts = len(vector)
        peaks = locmax(vector, indices=True)
        return self.spreadpeaks(zip(peaks, vector[peaks], strict=False),
                                npoints=npts, width=width)

    def spreadpeaks(
        self,
        peaks: Iterable[tuple[int, float]],
        npoints: int | None = None,
        width: float = 4.0,
        base: np.ndarray | None = None,
    ) -> np.ndarray:
        """ Generate a vector consisting of the max of a set of Gaussian bumps
        :params:
          peaks : list
            list of (index, value) pairs giving the center point and height
            of each gaussian
          npoints : int
            the length of the output vector (needed if base not provided)
          width : float
            the half-width of the Gaussians to lay down at each point
          base : np.array
            optional initial lower bound to place Gaussians above
        :returns:
          vector : np.array(npoints)
            the maximum across all the scaled Gaussians
        """
        if base is None:
            if npoints is None:
                raise ValueError("npoints must be provided when base is None")
            npoints_int = npoints
            vec = np.zeros(npoints_int)
        else:
            npoints_int = len(base)
            vec = np.copy(base)
        # binvals = np.arange(len(vec))
        # for pos, val in peaks:
        #   vec = np.maximum(vec, val*np.exp(-0.5*(((binvals - pos)
        #                                /float(width))**2)))
        if width != self.__sp_width or npoints_int != self.__sp_len:
            # Need to calculate new vector
            self.__sp_width = width
            self.__sp_len = npoints_int
            self.__sp_vals = np.exp(-0.5 * ((np.arange(-npoints_int, npoints_int + 1)
                                             / width)**2))
        # Now the actual function
        for pos, val in peaks:
            vec = np.maximum(vec, val * self.__sp_vals[np.arange(npoints_int)
                                                       + npoints_int - pos])
        return vec

    def _decaying_threshold_fwd_prune(self, sgram: np.ndarray, a_dec: float) -> np.ndarray:
        """ forward pass of findpeaks
            initial threshold envelope based on peaks in first 10 frames
        """
        (srows, scols) = np.shape(sgram)
        sthresh = self.spreadpeaksinvector(
            np.max(sgram[:, :np.minimum(10, scols)], axis=1), self.f_sd
        )
        # Store sthresh at each column, for debug
        # thr = np.zeros((srows, scols))
        peaks = np.zeros((srows, scols))
        # optimization of mask update
        __sp_pts = len(sthresh)
        __sp_v = self.__sp_vals

        for col in range(scols):
            s_col = sgram[:, col]
            # Find local magnitude peaks that are above threshold
            sdmaxposs = np.nonzero(locmax(s_col) * (s_col > sthresh))[0]
            # Work down list of peaks in order of their absolute value
            # above threshold
            valspeaks = sorted(zip(s_col[sdmaxposs], sdmaxposs, strict=False), reverse=True)
            for val, peakpos in valspeaks[:self.maxpksperframe]:
                # What we actually want
                # sthresh = spreadpeaks([(peakpos, s_col[peakpos])],
                #                      base=sthresh, width=f_sd)
                # Optimization - inline the core function within spreadpeaks
                sthresh = np.maximum(sthresh,
                                     val * __sp_v[(__sp_pts - peakpos):
                                                  (2 * __sp_pts - peakpos)])
                peaks[peakpos, col] = 1
            sthresh *= a_dec
        return peaks

    def _decaying_threshold_bwd_prune_peaks(
        self,
        sgram: np.ndarray,
        peaks: np.ndarray,
        a_dec: float,
    ) -> np.ndarray:
        """ backwards pass of findpeaks """
        scols = np.shape(sgram)[1]
        # Backwards filter to prune peaks
        sthresh = self.spreadpeaksinvector(sgram[:, -1], self.f_sd)
        for col in range(scols, 0, -1):
            pkposs = np.nonzero(peaks[:, col - 1])[0]
            peakvals = sgram[pkposs, col - 1]
            for val, peakpos in sorted(zip(peakvals, pkposs, strict=False), reverse=True):
                if val >= sthresh[peakpos]:
                    # Setup the threshold
                    sthresh = self.spreadpeaks([(peakpos, val)], base=sthresh,
                                               width=self.f_sd)
                    # Delete any following peak (threshold should, but be sure)
                    if col < scols:
                        peaks[peakpos, col] = 0
                else:
                    # delete the peak
                    peaks[peakpos, col - 1] = 0
            sthresh = a_dec * sthresh
        return peaks

    def find_peaks(self, d: np.ndarray, sr: int) -> list[tuple[int, int]]:
        """ Find the local peaks in the spectrogram as basis for fingerprints.
            Returns a list of (time_frame, freq_bin) pairs.

        :params:
          d - np.array of float
            Input waveform as 1D vector

          sr - int
            Sampling rate of d (not used)

        :returns:
          pklist - list of (int, int)
            Ordered list of landmark peaks found in STFT.  First value of
            each pair is the time index (in STFT frames, i.e., units of
            n_hop/sr secs), second is the FFT bin (in units of sr/n_fft
            Hz).
        """
        if len(d) == 0:
            return []

        # masking envelope decay constant
        a_dec = (1 - 0.01 * (self.density * np.sqrt(self.n_hop / 352.8) / 35)) ** (1 / OVERSAMP)
        # Take spectrogram
        mywin = np.hanning(self.n_fft + 2)[1:-1]
        sgram = np.abs(stft.stft(d, n_fft=self.n_fft,
                                 hop_length=self.n_hop,
                                 window=mywin))
        sgrammax = np.max(sgram)
        if sgrammax > 0.0:
            sgram = np.log(np.maximum(sgram, np.max(sgram) / 1e6))
            sgram = sgram - np.mean(sgram)
        else:
            # The sgram is identically zero, i.e., the input signal was identically
            # zero.  Not good, but let's let it through for now.
            logger.debug("find_peaks: Warning: input signal is identically zero.")
        # High-pass filter onset emphasis
        # [:-1,] discards top bin (nyquist) of sgram so bins fit in 8 bits
        sgram = np.array([scipy.signal.lfilter([1, -1],
                                               [1, -HPF_POLE ** (1 / OVERSAMP)], s_row)
                          for s_row in sgram])[:-1, ]
        # Prune to keep only local maxima in spectrum that appear above an online,
        # decaying threshold
        peaks = self._decaying_threshold_fwd_prune(sgram, a_dec)
        # Further prune these peaks working backwards in time, to remove small peaks
        # that are closely followed by a large peak
        peaks = self._decaying_threshold_bwd_prune_peaks(sgram, peaks, a_dec)
        # build a list of peaks we ended up with
        scols = np.shape(sgram)[1]
        pklist: list[tuple[int, int]] = []
        for col in range(scols):
            pklist.extend((col, bin_) for bin_ in np.nonzero(peaks[:, col])[0])
        return pklist

    def peaks2landmarks(
        self,
        pklist: Sequence[tuple[int, int]],
    ) -> list[tuple[int, int, int, int]]:
        """ Take a list of local peaks in spectrogram
            and form them into pairs as landmarks.
            pklist is a column-sorted list of (col, bin) pairs as created
            by findpeaks().
            Return a list of (col, peak, peak2, col2-col) landmark descriptors.
        """
        # Form pairs of peaks into landmarks
        landmarks: list[tuple[int, int, int, int]] = []
        if len(pklist) > 0:
            # Find column of the final peak in the list
            scols = pklist[-1][0] + 1
            # Convert (col, bin) list into peaks_at[col] lists
            peaks_at: list[list[int]] = [[] for _ in range(scols)]
            for (col, bin_) in pklist:
                peaks_at[col].append(bin_)

            # Build list of landmarks <starttime F1 endtime F2>
            for col in range(scols):
                for peak in peaks_at[col]:
                    pairsthispeak = 0
                    for col2 in range(col + self.mindt,
                                       min(scols, col + self.targetdt)):
                        if pairsthispeak < self.maxpairsperpeak:
                            for peak2 in peaks_at[col2]:
                                if abs(peak2 - peak) < self.targetdf and pairsthispeak < self.maxpairsperpeak:
                                    landmarks.append((col, peak,
                                                      peak2, col2 - col))
                                    pairsthispeak += 1

        return landmarks

    def wavfile2peaks(
        self,
        filename: str,
        shifts: int | None = None,
    ) -> list[tuple[int, int]] | list[list[tuple[int, int]]]:
        """ Read a soundfile and return its landmark peaks as a
            list of (time, bin) pairs.  If specified, resample to sr first.
            shifts > 1 causes hashes to be extracted from multiple shifts of
            waveform, to reduce frame effects.  """
        ext = os.path.splitext(filename)[1]
        peaks: list[tuple[int, int]] | list[list[tuple[int, int]]]
        if ext == PRECOMPPKEXT:
            # short-circuit - precomputed fingerprint file
            peaks = peaks_load(filename)
            dur = np.max(peaks, axis=0)[0] * self.n_hop / self.target_sr
        else:
            try:
                # [d, sr] = librosa.load(filename, sr=self.target_sr)
                d, sr = audio.audio_read(
                    filename,
                    sr=self.target_sr,
                    channels=1,
                    ffmpeg_path=self.ffmpeg_path,
                )
            except Exception as e:  # audioread.NoBackendError:
                message = f"wavfile2peaks: Error reading {filename}"
                if self.fail_on_error:
                    logger.exception(e)
                    raise IOError(message) from e
                logger.debug(message, "skipping")
                d = np.array([], dtype=float)
                sr = self.target_sr
            # Store duration in a global because it's hard to handle
            dur = len(d) / sr
            if shifts is None or shifts < 2:
                peaks = self.find_peaks(d, sr)
            else:
                # Calculate hashes with optional part-frame shifts
                peaklists: list[list[tuple[int, int]]] = []
                for shift in range(shifts):
                    shiftsamps = int(shift / self.shifts * self.n_hop)
                    peaklists.append(self.find_peaks(d[shiftsamps:], sr))
                peaks = peaklists

        # instrumentation to track total amount of sound processed
        self.soundfiledur = dur
        self.soundfiletotaldur += dur
        self.soundfilecount += 1
        return peaks

    def wavfile2hashes(self, filename: str) -> np.ndarray:
        """ Read a soundfile and return its fingerprint hashes as a
            list of (time, hash) pairs.  If specified, resample to sr first.
            shifts > 1 causes hashes to be extracted from multiple shifts of
            waveform, to reduce frame effects.  """
        ext = os.path.splitext(filename)[1]
        if ext == PRECOMPEXT:
            # short-circuit - precomputed fingerprint file
            hashes = np.array(hashes_load(filename), dtype=np.int32)
            dur = np.max(hashes, axis=0)[0] * self.n_hop / self.target_sr
            # instrumentation to track total amount of sound processed
            self.soundfiledur = dur
            self.soundfiletotaldur += dur
            self.soundfilecount += 1
        else:
            peaks = self.wavfile2peaks(filename, self.shifts)
            if len(peaks) == 0:
                return np.zeros((0, 2), dtype=np.int32)
            # Did we get returned a list of lists of peaks due to shift?
            if isinstance(peaks[0], list):
                peaklists = cast(list[list[tuple[int, int]]], peaks)
                peak_arrays = [
                    landmarks2hashes(self.peaks2landmarks(peaklist))
                    for peaklist in peaklists
                ]
                query_hashes = (
                    np.concatenate(peak_arrays)
                    if len(peak_arrays) > 0
                    else np.zeros((0, 2), dtype=np.int32)
                )
            else:
                peak_pairs = cast(list[tuple[int, int]], peaks)
                query_hashes = landmarks2hashes(self.peaks2landmarks(peak_pairs))

            # Remove duplicates by merging each row into a single value.
            hashes_hashes = (((query_hashes[:, 0].astype(np.uint64)) << 32)
                             + query_hashes[:, 1].astype(np.uint64))
            unique_hash_hash = np.sort(np.unique(hashes_hashes))
            unique_hashes = np.hstack([
                (unique_hash_hash >> 32)[:, np.newaxis],
                (unique_hash_hash & ((1 << 32) - 1))[:, np.newaxis]
            ]).astype(np.int32)
            hashes = unique_hashes
                # Or simply np.unique(query_hashes, axis=0) for numpy >= 1.13

        # print("wavfile2hashes: read", len(hashes), "hashes from", filename)
        return hashes

    # ########## functions to link to actual hash table index database ###### #

    def ingest(self, hashtable: hash_table.HashTable, filename: str) -> tuple[float, int]:
        """ Read an audio file and add it to the database
        :params:
          hashtable : HashTable object
            the hash table to add to
          filename : str
            name of the soundfile to add
        :returns:
          dur : float
            the duration of the track
          nhashes : int
            the number of hashes it mapped into
        """
        # sr = 11025
        # print("ingest: sr=",sr)
        # d, sr = librosa.load(filename, sr=sr)
        # librosa.load on mp3 files prepends 396 samples compared
        # to Matlab audioread ??
        # hashes = landmarks2hashes(peaks2landmarks(find_peaks(d, sr,
        #                                                     density=density,
        #                                                     n_fft=n_fft,
        #                                                     n_hop=n_hop)))
        hashes = self.wavfile2hashes(filename)
        hashtable.store(
            filename,
            [(int(time_), int(hash_)) for time_, hash_ in hashes.tolist()],
        )
        # return (len(d)/float(sr), len(hashes))
        # return (np.max(hashes, axis=0)[0]*n_hop/float(sr), len(hashes))
        # soundfiledur is set up in wavfile2hashes, use result here
        return self.soundfiledur, len(hashes)


# ########## functions to read/write hashes to file for a single track #### #

# Format string for writing binary data to file
HASH_FMT = '<2i'
HASH_MAGIC = b'audfprinthashV00'  # 16 chars, FWIW
PEAK_FMT = '<2i'
PEAK_MAGIC = b'audfprintpeakV00'  # 16 chars, FWIW


def hashes_save(hashfilename: str, hashes: Iterable[tuple[int, int]]) -> None:
    """ Write out a list of (time, hash) pairs as 32 bit ints """
    with open(hashfilename, 'wb') as f:
        f.write(HASH_MAGIC)
        for time_, hash_ in hashes:
            f.write(struct.pack(HASH_FMT, time_, hash_))


def hashes_load(hashfilename: str) -> list[tuple[int, int]]:
    """ Read back a set of hashes written by hashes_save """
    hashes = []
    fmtsize = struct.calcsize(HASH_FMT)
    with open(hashfilename, 'rb') as f:
        magic = f.read(len(HASH_MAGIC))
        if magic != HASH_MAGIC:
            raise IOError(f'{hashfilename} is not a hash file (magic {magic!r})')
        data = f.read(fmtsize)
        while data is not None and len(data) == fmtsize:
            hashes.append(struct.unpack(HASH_FMT, data))
            data = f.read(fmtsize)
    return hashes


def peaks_save(peakfilename: str, peaks: Iterable[tuple[int, int]]) -> None:
    """ Write out a list of (time, bin) pairs as 32 bit ints """
    with open(peakfilename, 'wb') as f:
        f.write(PEAK_MAGIC)
        for time_, bin_ in peaks:
            f.write(struct.pack(PEAK_FMT, time_, bin_))


def peaks_load(peakfilename: str) -> list[tuple[int, int]]:
    """ Read back a set of (time, bin) pairs written by peaks_save """
    peaks = []
    fmtsize = struct.calcsize(PEAK_FMT)
    with open(peakfilename, 'rb') as f:
        magic = f.read(len(PEAK_MAGIC))
        if magic != PEAK_MAGIC:
            raise IOError(f'{peakfilename} is not a peak file (magic {magic!r})')
        data = f.read(fmtsize)
        while data is not None and len(data) == fmtsize:
            peaks.append(struct.unpack(PEAK_FMT, data))
            data = f.read(fmtsize)
    return peaks


# ####### function signature for Gordon feature extraction
# ####### which stores the precalculated hashes for each track separately

extract_features_analyzer: Analyzer | None = None


def extract_features(track_obj: Any, *args: Any, **kwargs: Any) -> np.ndarray:
    """ Extract the audfprint fingerprint hashes for one file.
    :params:
      track_obj : object
        Gordon's internal structure defining a track; we use
        track_obj.fn_audio to find the actual audio file.
    :returns:
      hashes : list of (int, int)
        The times (in frames) and hashes analyzed from the audio file.
    """
    global extract_features_analyzer
    if extract_features_analyzer is None:
        extract_features_analyzer = Analyzer()

    density = kwargs.get("density", None)
    if density is not None:
        extract_features_analyzer.density = float(density)
    n_fft = kwargs.get("n_fft", None)
    if n_fft is not None:
        extract_features_analyzer.n_fft = int(n_fft)
    n_hop = kwargs.get("n_hop", None)
    if n_hop is not None:
        extract_features_analyzer.n_hop = int(n_hop)
    sr = kwargs.get("sr", None)
    if sr is not None:
        extract_features_analyzer.target_sr = int(sr)
    return extract_features_analyzer.wavfile2hashes(track_obj.fn_audio)


# Handy function to build a new hash table from a file glob pattern
g2h_analyzer: Analyzer | None = None


def glob2hashtable(pattern: str, density: float = 20.0) -> hash_table.HashTable:
    """ Build a hash table from the files matching a glob pattern """
    global g2h_analyzer
    if g2h_analyzer is None:
        g2h_analyzer = Analyzer(density=density)

    ht = hash_table.HashTable()
    filelist = glob.glob(pattern)
    initticks = time.perf_counter()
    totdur = 0.0
    tothashes = 0
    for ix, file_ in enumerate(filelist):
        logger.debug(time.ctime(), "ingesting #", ix, ":", file_, "...")
        dur, nhash = g2h_analyzer.ingest(ht, file_)
        totdur += dur
        tothashes += nhash
    elapsedtime = time.perf_counter() - initticks
    logger.debug("Added", tothashes, "(", tothashes / totdur, "hashes/sec) at ",
          elapsedtime / totdur, "x RT")
    return ht


def local_tester() -> None:
    test_fn = '/Users/dpwe/Downloads/carol11k.wav'
    test_ht = hash_table.HashTable()
    test_analyzer = Analyzer()

    test_analyzer.ingest(test_ht, test_fn)
    test_ht.save('httest.pklz')


# Run the test function if called from the command line
if __name__ == "__main__":
    local_tester()



# coding=utf-8
"""
hash_table.py

Python implementation of the very simple, fixed-array hash table
used for the audfprint2 fingerprinter.

2014-05-25 Dan Ellis dpwe@ee.columbia.edu
"""

import gzip as zlib
import json
import logging
import math
import os
import pickle
import random
from collections.abc import Callable, Iterable, Sequence
from enum import Enum
from typing import IO, Any

import h5py  # type: ignore[import-untyped]
import numpy as np
import scipy.io  # type: ignore[import-untyped]

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger("audfprint")

basestring = (str, bytes)
pickle_options = {"encoding": "latin1"}


class DatabaseType(Enum):
    PKL = "PKL"
    HDF = "HDF"
    MAT = "MAT"


# Current format version
HT_VERSION = 20170724
# Earliest acceptable version
HT_COMPAT_VERSION = 20170724
# Earliest version that can be updated with load_old
HT_OLD_COMPAT_VERSION = 20140920


def _bitsfor(maxval: int) -> int:
    """Convert a maxval into a number of bits (left shift).
    Raises a ValueError if the maxval is not a power of 2."""
    maxvalbits = int(round(math.log(maxval) / math.log(2)))
    if maxval != (1 << maxvalbits):
        raise ValueError("maxval must be a power of 2, not %d" % maxval)
    return maxvalbits


class HashTable(object):
    """
    Simple hash table for storing and retrieving fingerprint hashes.

    :usage:
       >>> ht = HashTable(size=2**10, depth=100)
       >>> ht.store('identifier', list_of_landmark_time_hash_pairs)
       >>> list_of_ids_tracks = ht.get_hits(hash)
    """

    __slots__ = (
        "hashbits",
        "depth",
        "maxtimebits",
        "table",
        "counts",
        "names",
        "hashesperid",
        "params",
        "ht_version",
        "dirty",
    )

    def __init__(
        self,
        filename: str | None = None,
        hashbits: int = 20,
        depth: int = 100,
        maxtime: int = 16384,
    ) -> None:
        """allocate an empty hash table of the specified size"""
        if filename is not None:
            self.load(filename)
        else:
            self.hashbits = hashbits
            self.depth = depth
            self.maxtimebits = _bitsfor(maxtime)
            # allocate the big table
            size = 2**hashbits
            self.table = np.zeros((size, depth), dtype=np.uint32)
            # keep track of number of entries in each list
            self.counts = np.zeros(size, dtype=np.int32)
            # map names to IDs
            self.names: list[str | None] = []
            # track number of hashes stored per id
            self.hashesperid = np.zeros(0, np.uint32)
            # Empty params
            self.params: dict[str, Any] = {}
            # Record the current version
            self.ht_version = HT_VERSION
            # Mark as unsaved
            self.dirty = True

    def reset(self) -> None:
        """Reset to empty state (but preserve parameters)"""
        self.table[:, :] = 0
        self.counts[:] = 0
        self.names = []
        self.hashesperid.resize(0)
        self.dirty = True

    def store(self, name: str | int, timehashpairs: Sequence[tuple[int, int]]) -> None:
        """Store a list of hashes in the hash table
        associated with a particular name (or integer ID) and time.
        """
        id_ = self.name_to_id(name, add_if_missing=True)
        # Now insert the hashes
        hashmask = (1 << self.hashbits) - 1
        # mxtime = self.maxtime
        maxtime = 1 << self.maxtimebits
        timemask = maxtime - 1
        # Try sorting the pairs by hash value, for better locality in storing
        # sortedpairs = sorted(timehashpairs, key=lambda x:x[1])
        sortedpairs = timehashpairs
        # Tried making it an np array to permit vectorization, but slower...
        # sortedpairs = np.array(sorted(timehashpairs, key=lambda x:x[1]),
        #                       dtype=int)
        # Keep only the bottom part of the time value
        # sortedpairs[:,0] = sortedpairs[:,0] % self.maxtime
        # Keep only the bottom part of the hash value
        # sortedpairs[:,1] = sortedpairs[:,1] & hashmask
        # The id value is based on (id_ + 1) to avoid an all-zero value.
        idval = (id_ + 1) << self.maxtimebits
        for time_, hash_ in sortedpairs:
            # Keep only the bottom part of the hash value
            hash_ &= hashmask
            # How many already stored for this hash?
            count = self.counts[hash_]
            # Keep only the bottom part of the time value
            # time_ %= mxtime
            time_ &= timemask
            # Mixin with ID
            val = idval + time_  # .astype(np.uint32)
            if count < self.depth:
                # insert new val in next empty slot
                # slot = self.counts[hash_]
                self.table[hash_, count] = val
            else:
                # Choose a point at random
                slot = random.randint(0, count)
                # Only store if random slot wasn't beyond end
                if slot < self.depth:
                    self.table[hash_, slot] = val
            # Update record of number of vals in this bucket
            self.counts[hash_] = count + 1
        # Record how many hashes we (attempted to) save for this id
        self.hashesperid[id_] += len(timehashpairs)
        # Mark as unsaved
        self.dirty = True

    def get_entry(self, hash_: int) -> np.ndarray:
        """Return np.array of [id, time] entries
        associate with the given hash as rows.
        """
        vals = self.table[hash_, : min(self.depth, self.counts[hash_])]
        maxtimemask = (1 << self.maxtimebits) - 1
        # ids we report externally start at 0, but in table they start at 1.
        ids = (vals >> self.maxtimebits) - 1
        return np.c_[ids, vals & maxtimemask].astype(np.int32)

    def get_hits(self, hashes: np.ndarray) -> np.ndarray:
        """Return np.array of [id, delta_time, hash, time] rows
        associated with each element in hashes array of [time, hash] rows.
        This version has get_entry() inlined, it's about 30% faster.
        """
        # Allocate to largest possible number of hits
        nhashes = np.shape(hashes)[0]
        hits = np.zeros((nhashes * self.depth, 4), np.int32)
        nhits = 0
        maxtimemask = (1 << self.maxtimebits) - 1
        hashmask = (1 << self.hashbits) - 1
        # Fill in
        for ix in range(nhashes):
            time_ = hashes[ix][0]
            hash_ = hashmask & hashes[ix][1]
            nids = min(self.depth, self.counts[hash_])
            tabvals = self.table[hash_, :nids]
            hitrows = nhits + np.arange(nids)
            # Make external IDs start from 0.
            hits[hitrows, 0] = (tabvals >> self.maxtimebits) - 1
            hits[hitrows, 1] = (tabvals & maxtimemask) - time_
            hits[hitrows, 2] = hash_
            hits[hitrows, 3] = time_
            nhits += nids
        # Discard the excess rows
        hits.resize((nhits, 4), refcheck=False)
        return hits

    def save(
        self,
        name: str,
        params: dict | None = None,
        file_object: IO[bytes] | None = None,
        save_type: DatabaseType | str | None = None,
    ) -> str:
        base, ext = os.path.splitext(name)
        ext = ext.lower()

        if not save_type:
            if ext == ".hdf" or ext not in (".pkl", ".pklz"):
                save_type = DatabaseType.HDF
            else:
                save_type = DatabaseType.PKL
        if save_type in [DatabaseType.HDF, DatabaseType.HDF.value]:
            if ext != ".hdf":
                name = f"{base}.hdf"
            self.save_hdf(name)
        elif save_type in [DatabaseType.PKL, DatabaseType.PKL.value]:
            if ext not in (".pkl", ".pklz"):
                name = f"{base}.pklz"
            self.save_pkl(name)
        else:
            raise ValueError(f"Unknown database type or doesn't support to export: {save_type}")

        self.dirty = False
        self._extracted_from_load_28("Saved fprints for ", " hashes) to ", name)
        return name

    def save_pkl(
        self,
        name: str,
        params: dict | None = None,
        file_object: IO[bytes] | None = None,
    ) -> None:
        """Save hash table to file <name>,
        including optional addition params
        """
        # Merge in any provided params
        if params:
            for key in params:
                self.params[key] = params[key]
        f = file_object or zlib.open(name, "wb")
        pickle.dump(self, f, pickle.HIGHEST_PROTOCOL)

    def save_hdf(
        self,
        name: str,
        params: dict | None = None,
        file_object: h5py.File | None = None,
    ) -> None:
        """Save hash table to file <name>,
        including optional additional params
        """
        # Merge in any provided params
        if params:
            for key in params:
                self.params[key] = params[key]

        f = file_object or name
        temp = h5py.File(f, "w")

        temp.attrs["params"] = json.dumps(self.params)
        temp.attrs["hashbits"] = self.hashbits
        temp.attrs["depth"] = self.depth
        temp.attrs["maxtimebits"] = self.maxtimebits

        temp.create_dataset("table", data=self.table, compression="gzip", compression_opts=1)
        temp.create_dataset("counts", data=self.counts)
        temp.create_dataset("names", data=self.names)
        temp.create_dataset("hashesperid", data=self.hashesperid)

        # temp.close()

    def load(self, name: str) -> None:
        """Read either pklz or mat-format hash table file"""
        logger.debug(f"Loading hash table from {name}")

        ext = os.path.splitext(name)[1]
        logger.debug(f"File extension is {ext}")
        if ext == ".mat":
            self.load_matlab(name)
        elif ext == ".hdf":
            self.load_hdf(name)
        elif ext in [".pkl", ".pklz"]:
            self.load_pkl(name)
        else:
            logger.debug("Hash table file type is not specified. Loading as HDF")
            self.load_hdf(name)

        self._extracted_from_load_28("Read fprints for ", " hashes) from ", name)

    # TODO Rename this here and in `save` and `load`
    def _extracted_from_load_28(self, arg0: str, arg1: str, name: str) -> None:
        nhashes = sum(self.counts)
        dropped = nhashes - sum(np.minimum(self.depth, self.counts))
        num_files = sum(n is not None for n in self.names)
        drop_percent = 100.0 * dropped / max(1, nhashes)
        logger.debug(
            f"{arg0}{num_files} files ({nhashes}{arg1}{name} ({drop_percent:.2f}% dropped)"
        )

    def load_hdf(
        self,
        name: str,
        file_object: h5py.File | None = None,
    ) -> None:
        """Read hash table values from pickle file <name>."""
        f = file_object or name
        temp = h5py.File(f, "r")

        # if temp.ht_version < HT_OLD_COMPAT_VERSION:
        #     raise ValueError('Version of ' + name + ' is ' + str(temp.ht_version)
        #                      + ' which is not at least ' +
        #                      str(HT_OLD_COMPAT_VERSION))

        # assert temp.ht_version >= HT_COMPAT_VERSION

        self.hashbits = temp.attrs["hashbits"]
        self.depth = temp.attrs["depth"]

        if "maxtimebits" in temp.attrs:
            self.maxtimebits = temp.attrs["maxtimebits"]
        else:
            raise ValueError("'maxtimebits' not found!")
            # self.maxtimebits = _bitsfor(temp.maxtime)

        self.table = temp["table"][:]

        self.counts = temp["counts"][:]
        self.names = list(temp["names"].asstr())
        self.hashesperid = np.array(temp["hashesperid"][...]).astype(np.uint32)
        self.dirty = False
        self.params = json.loads(temp.attrs["params"])

        # self.ht_version = temp.ht_version

    def load_pkl(
        self,
        name: str,
        file_object: IO[bytes] | None = None,
    ) -> None:
        """Read hash table values from pickle file <name>."""
        f = file_object or zlib.open(name, "rb")
        temp = pickle.load(f, encoding=pickle_options["encoding"])
        if temp.ht_version < HT_OLD_COMPAT_VERSION:
            raise ValueError(
                f"Version of {name} is {str(temp.ht_version)} which is not at least {str(HT_OLD_COMPAT_VERSION)}"
            )
        # assert temp.ht_version >= HT_COMPAT_VERSION
        self.hashbits = temp.hashbits
        self.depth = temp.depth
        if hasattr(temp, "maxtimebits"):
            self.maxtimebits = temp.maxtimebits
        else:
            self.maxtimebits = _bitsfor(temp.maxtime)
        if temp.ht_version < HT_COMPAT_VERSION:
            # Need to upgrade the database.
            logger.debug("Loading database version", temp.ht_version, "in compatibility mode.")
            # Offset all the nonzero bins with one ID count.
            temp.table += np.array(1 << self.maxtimebits).astype(np.uint32) * (temp.table != 0)
            temp.ht_version = HT_VERSION

        self.table = temp.table
        self.ht_version = temp.ht_version
        self.counts = temp.counts
        self.names = temp.names
        self.hashesperid = np.array(temp.hashesperid).astype(np.uint32)
        self.dirty = False
        self.params = temp.params

    def load_matlab(self, name: str) -> None:
        """Read hash table from version saved by Matlab audfprint.
        :params:
          name : str
            filename of .mat matlab fp dbase file
        :side_effects:
          Sets up attributes of self including
          params : dict
            dictionary of parameters from the Matlab file including
              'mat_version' : float
                version read from Matlab file (must be >= 0.90)
              'hoptime' : float
                hoptime read from Matlab file (must be 0.02322)
              'targetsr' : float
                target sampling rate from Matlab file (must be 11025)
        """
        mht = scipy.io.loadmat(name)
        params = {"mat_version": mht["HT_params"][0][0][-1][0][0]}
        assert params["mat_version"] >= 0.9
        self.hashbits = _bitsfor(mht["HT_params"][0][0][0][0][0])
        self.depth = mht["HT_params"][0][0][1][0][0]
        self.maxtimebits = _bitsfor(mht["HT_params"][0][0][2][0][0])
        params["hoptime"] = mht["HT_params"][0][0][3][0][0]
        params["targetsr"] = mht["HT_params"][0][0][4][0][0]
        params["nojenkins"] = mht["HT_params"][0][0][5][0][0]
        # Python doesn't support the (pointless?) jenkins hashing
        assert params["nojenkins"]
        self.table = mht["HashTable"].T
        self.counts = mht["HashTableCounts"][0]
        self.names = [str(val[0]) if len(val) > 0 else None for val in mht["HashTableNames"][0]]
        self.hashesperid = np.array(mht["HashTableLengths"][0]).astype(np.uint32)
        # Matlab uses 1-origin for the IDs in the hashes, but the Python code
        # also skips using id_ 0, so that names[0] corresponds to id_ 1.
        # Otherwise unmodified database
        self.dirty = False
        self.params = params

    def totalhashes(self) -> int:
        """Return the total count of hashes stored in the table"""
        return int(np.sum(self.counts))

    def merge(self, ht: "HashTable") -> None:
        """Merge in the results from another hash table"""
        # All the items go into our table, offset by our current size
        # Check compatibility
        assert self.maxtimebits == ht.maxtimebits
        ncurrent = len(self.names)
        # size = len(self.counts)
        self.names += ht.names
        self.hashesperid = np.append(self.hashesperid, ht.hashesperid)
        # Shift all the IDs in the second table down by ncurrent
        idoffset = (1 << self.maxtimebits) * ncurrent
        for hash_ in np.nonzero(ht.counts)[0]:
            allvals = np.r_[self.table[hash_, : self.counts[hash_]], ht.table[hash_, : ht.counts[hash_]] + idoffset]
            # ht.counts[hash_] may be more than the actual number of
            # hashes we obtained, if ht.counts[hash_] > ht.depth.
            # Subselect based on actual size.
            if len(allvals) > self.depth:
                # Our hash bin is filled: randomly subselect the
                # hashes, and update count to accurately track the
                # total number of hashes we've seen for this bin.
                somevals = np.random.permutation(allvals)[: self.depth]
                self.table[hash_,] = somevals
                self.counts[hash_] += ht.counts[hash_]
            else:
                # Our bin isn't full.  Store all the hashes, and
                # accurately track how many values it contains.  This
                # may mean some of the hashes counted for full buckets
                # in ht are "forgotten" if ht.depth < self.depth.
                self.table[hash_, : len(allvals)] = allvals
                self.counts[hash_] = len(allvals)

        self.dirty = True

    def name_to_id(self, name: str | int, add_if_missing: bool = False) -> int:
        """Lookup name in the names list, or optionally add."""
        if isinstance(name, basestring):
            # lookup name or assign new
            if name not in self.names:
                if not add_if_missing:
                    raise ValueError(f"name {name} not found")
                # Use an empty slot in the list if one exists.
                try:
                    id_ = self.names.index(None)
                    self.names[id_] = name
                    self.hashesperid[id_] = 0
                except ValueError:
                    self.names.append(name)
                    self.hashesperid = np.append(self.hashesperid, [0])
            id_ = self.names.index(name)
        else:
            # we were passed in a numerical id
            id_ = name
        return id_

    def remove(self, name: str | int) -> None:
        """Remove all data for named entity from the hash table."""
        id_ = self.name_to_id(name)
        # Top nybbles of table entries are id_ + 1 (to avoid all-zero entries)
        id_in_table = (self.table >> self.maxtimebits) == id_ + 1
        hashes_removed = 0
        for hash_ in np.nonzero(np.max(id_in_table, axis=1))[0]:
            vals = self.table[hash_, : self.counts[hash_]]
            filtered_vals = [v for v, x in zip(vals, id_in_table[hash_], strict=False) if not x]
            vals_array = np.array(filtered_vals)
            self.table[hash_] = np.hstack([vals_array, np.zeros(self.depth - len(vals_array))])
            # This will forget how many extra hashes we had dropped until now.
            self.counts[hash_] = len(filtered_vals)
            hashes_removed += np.sum(id_in_table[hash_])
        self.names[id_] = None
        self.hashesperid[id_] = 0
        self.dirty = True
        logger.debug("Removed", name, "(", hashes_removed, "hashes).")

    def retrieve(self, name: str | int) -> np.ndarray:
        """Return an np.array of (time, hash) pairs found in the table."""
        id_ = self.name_to_id(name)
        maxtimemask = (1 << self.maxtimebits) - 1
        num_hashes_per_hash = np.sum((self.table >> self.maxtimebits) == (id_ + 1), axis=1)
        hashes_containing_id = np.nonzero(num_hashes_per_hash)[0]
        timehashpairs = np.zeros((sum(num_hashes_per_hash), 2), dtype=np.int32)
        hashes_so_far = 0
        for hash_ in hashes_containing_id:
            entries = self.table[hash_, : self.counts[hash_]]
            matching_entries = np.nonzero((entries >> self.maxtimebits) == (id_ + 1))[0]
            times = entries[matching_entries] & maxtimemask
            timehashpairs[hashes_so_far : hashes_so_far + len(times), 0] = times
            timehashpairs[hashes_so_far : hashes_so_far + len(times), 1] = hash_
            hashes_so_far += len(times)
        return timehashpairs

    def list(self, print_fn: Callable[[str], None] | None = None) -> None:
        """List all the known items."""
        if not print_fn:
            print_fn = print
        for name, count in zip(self.names, self.hashesperid, strict=False):
            if name:
                print_fn(f"{name} ({str(count)} hashes)")



# coding=utf-8
"""
audfprint2_match.py

Fingerprint matching code for audfprint2

2014-05-26 Dan Ellis dpwe@ee.columbia.edu
"""

import logging
import os
import time
from typing import cast

import numpy as np
import psutil  # type: ignore[import-untyped]
import scipy.signal  # type: ignore[import-untyped]

from audfprint2.core import analyzer, hash_table
from audfprint2.utils import audio, stft

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger("audfprint2")
TRACE_LEVEL = logging.DEBUG - 5


def process_info() -> tuple[int, float]:
    rss = usrtime = 0
    p = psutil.Process(os.getpid())
    if os.name == 'nt':
        rss = p.memory_info()[0]
        usrtime = p.cpu_times()[0]
    else:
        rss = p.get_memory_info()[0]
        usrtime = p.get_cpu_times()[0]
    return rss, usrtime


def log(message: str) -> None:
    """ log info with stats """
    rss, usrtime = process_info()
    logger.debug(f'{time.ctime()} physmem={rss} utime={usrtime} {message}')


def encpowerof2(val: int) -> int:
    """ Return N s.t. 2^N >= val """
    return int(np.ceil(np.log(max(1, val)) / np.log(2)))


def locmax(vec: np.ndarray, indices: bool = False) -> np.ndarray:
    """ Return a boolean vector of which points in vec are local maxima.
        End points are peaks if larger than single neighbors.
        if indices=True, return the indices of the True values instead
        of the boolean vector. (originally from audfprint2.py)
    """
    # x[-1]-1 means last value can be a peak
    # nbr = np.greater_equal(np.r_[x, x[-1]-1], np.r_[x[0], x])
    # the np.r_ was killing us, so try an optimization...
    nbr = np.zeros(len(vec) + 1, dtype=bool)
    nbr[0] = True
    nbr[1:-1] = np.greater_equal(vec[1:], vec[:-1])
    maxmask = (nbr[:-1] & ~nbr[1:])
    return np.nonzero(maxmask)[0] if indices else maxmask


def keep_local_maxes(vec: np.ndarray) -> np.ndarray:
    """ Zero out values unless they are local maxima."""
    local_maxes = np.zeros(vec.shape)
    locmaxindices = locmax(vec, indices=True)
    local_maxes[locmaxindices] = vec[locmaxindices]
    return local_maxes


def find_modes(
    data: np.ndarray,
    threshold: int = 5,
    window: int = 0,
) -> tuple[np.ndarray, np.ndarray]:
    """ Find multiple modes in data,  Report a list of (mode, count)
        pairs for every mode greater than or equal to threshold.
        Only local maxima in counts are returned.
    """
    # TODO: Ignores window at present
    datamin = np.amin(data)
    fullvector = np.bincount(data - datamin)
    # Find local maxima
    localmaxes = np.nonzero(np.logical_and(locmax(fullvector),
                                           np.greater_equal(fullvector,
                                                            threshold)))[0]
    return localmaxes + datamin, fullvector[localmaxes]


class Matcher(object):
    """Provide matching for audfprint2 fingerprint queries to hash table"""

    def __init__(self) -> None:
        """Set up default object values"""
        # Tolerance window for time differences
        self.window = 1
        # Absolute minimum number of matching hashes to count as a match
        self.threshcount = 5
        # How many hits to return?
        self.max_returns = 1
        # How deep to search in return list?
        self.search_depth = 100
        # Sort those returns by time (instead of counts)?
        self.sort_by_time = False
        # Do illustration?
        self.illustrate = False
        # Careful counts?
        self.exact_count = False
        # Search for time range?
        self.find_time_range = False
        # Quantile of time range to report.
        self.time_quantile = 0.02
        # Display pre-emphasized spectrogram in illustrate_match?
        self.illustrate_hpf = False
        # If there are a lot of matches within a single track at different
        # alignments, stop looking after a while.
        self.max_alignments_per_id = 100

    def _best_count_ids(
        self,
        hits: np.ndarray,
        ht: hash_table.HashTable,
    ) -> tuple[np.ndarray, np.ndarray]:
        """ Return the indexes for the ids with the best counts.
            hits is a matrix as returned by hash_table.get_hits()
            with rows of consisting of [id dtime hash otime] """
        allids = hits[:, 0]
        ids = np.unique(allids)
        # rawcounts = np.sum(np.equal.outer(ids, allids), axis=1)
        # much faster, and doesn't explode memory
        rawcounts = np.bincount(allids)[ids]
        # Divide the raw counts by the total number of hashes stored
        # for the ref track, to downweight large numbers of chance
        # matches against longer reference tracks.
        wtdcounts = rawcounts / (ht.hashesperid[ids].astype(float))

        # Find all the actual hits for a the most popular ids
        bestcountsixs = np.argsort(wtdcounts)[::-1]
        # We will examine however many hits have rawcounts above threshold
        # up to a maximum of search_depth.
        maxdepth = np.minimum(np.count_nonzero(np.greater(rawcounts,
                                                          self.threshcount)),
                              self.search_depth)
        # Return the ids to check
        bestcountsixs = bestcountsixs[:maxdepth]
        return ids[bestcountsixs], rawcounts[bestcountsixs]

    def _unique_match_hashes(self, id: int, hits: np.ndarray, mode: int) -> np.ndarray:
        """ Return the list of unique matching hashes.  Split out so
            we can recover the actual matching hashes for the best
            match if required. """
        allids = hits[:, 0]
        alltimes = hits[:, 1]
        allhashes = hits[:, 2].astype(np.int64)
        allotimes = hits[:, 3]
        timebits = max(1, encpowerof2(np.amax(allotimes)))
        # matchhashes may include repeats because multiple
        # ref hashes may match a single query hash under window.
        # Uniqify:
        # matchhashes = sorted(list(set(matchhashes)))
        # much, much faster:
        matchix = np.nonzero(
                np.logical_and(allids == id, np.less_equal(np.abs(alltimes - mode),
                                                           self.window)))[0]
        matchhasheshash = np.unique(allotimes[matchix]
                                    + (allhashes[matchix] << timebits))
        timemask = (1 << timebits) - 1
        return np.c_[matchhasheshash & timemask, matchhasheshash >> timebits]

    def _calculate_time_ranges(
        self,
        hits: np.ndarray,
        id: int,
        mode: int,
    ) -> tuple[int, int]:
        """Given the id and mode, return the actual time support.
           hits is an np.array of id, skew_time, hash, orig_time
           which must be sorted in orig_time order."""
        minoffset = mode - self.window
        maxoffset = mode + self.window
        # match_times = sorted(hits[row, 3]
        #                     for row in np.nonzero(hits[:, 0]==id)[0]
        #                     if mode - self.window <= hits[row, 1]
        #                     and hits[row, 1] <= mode + self.window)
        match_times = hits[np.logical_and.reduce([
            hits[:, 1] >= minoffset,
            hits[:, 1] <= maxoffset,
            hits[:, 0] == id
        ]), 3]
        min_time = match_times[int(len(match_times) * self.time_quantile)]
        max_time = match_times[int(len(match_times) * (1.0 - self.time_quantile)) - 1]
        # log("_calc_time_ranges: len(hits)={:d} id={:d} mode={:d} matches={:d} min={:d} max={:d}".format(
        #    len(hits), id, mode, np.sum(np.logical_and(hits[:, 1] >= minoffset,
        #                                               hits[:, 1] <= maxoffset)),
        #    min_time, max_time))
        return min_time, max_time

    def _exact_match_counts(
        self,
        hits: np.ndarray,
        ids: np.ndarray,
        rawcounts: np.ndarray,
        hashesfor: int | None = None,
    ) -> np.ndarray:
        """Find the number of "filtered" (time-consistent) matching hashes
            for each of the promising ids in <ids>.  Return an
            np.array whose rows are [id, filtered_count,
            modal_time_skew, unfiltered_count, original_rank,
            min_time, max_time].  Results are sorted by original rank
            (but will not in general include all the the original
            IDs).  There can be multiple rows for a single ID, if
            there are several distinct time_skews giving good
            matches.
        """
        # Sort hits into time_in_original order - needed for _calc_time_range
        sorted_hits = hits[hits[:, 3].argsort()]
        # Slower, old process for exact match counts
        allids = sorted_hits[:, 0]
        alltimes = sorted_hits[:, 1]
        # allhashes = sorted_hits[:, 2]
        # allotimes = sorted_hits[:, 3]
        # Allocate enough space initially for 4 modes per hit
        maxnresults = len(ids) * 4
        results = np.zeros((maxnresults, 7), np.int32)
        nresults = 0
        min_time = 0
        max_time = 0
        for urank, (id, rawcount) in enumerate(zip(ids, rawcounts, strict=False)):
            modes, counts = find_modes(alltimes[np.nonzero(allids == id)[0]],
                                       window=self.window,
                                       threshold=self.threshcount)
            for mode in modes:
                matchhashes = self._unique_match_hashes(id, sorted_hits, mode)
                # Now we get the exact count
                filtcount = len(matchhashes)
                if filtcount >= self.threshcount:
                    if nresults == maxnresults:
                        # Extend array
                        maxnresults *= 2
                        results.resize((maxnresults, results.shape[1]))
                    if self.find_time_range:
                        min_time, max_time = self._calculate_time_ranges(
                                sorted_hits, id, mode)
                    results[nresults, :] = [id, filtcount, mode, rawcount,
                                            urank, min_time, max_time]
                    nresults += 1
        return results[:nresults, :]

    def _approx_match_counts(
        self,
        hits: np.ndarray,
        ids: np.ndarray,
        rawcounts: np.ndarray,
    ) -> np.ndarray:
        """ Quick and slightly inaccurate routine to count time-aligned hits.

        Only considers largest mode for reference ID match.

        Args:
          hits: np.array of hash matches, each row consists of
            <track_id, skew_time, hash, orig_time>.
          ids: list of the IDs to check, based on raw match count.
          rawcounts: list giving the actual raw counts for each id to try.

        Returns:
            Rows of [id, filt_count, time_skew, raw_count, orig_rank,
            min_time, max_time].
            Ids occur in the same order as the input list, but ordering
            of (potentially multiple) hits within each track may not be
            sorted (they are sorted by the largest single count value, not
            the total count integrated over -window:+window bins).
        """
        # In fact, the counts should be the same as exact_match_counts
        # *but* some matches may be pruned because we don't bother to
        # apply the window (allowable drift in time alignment) unless
        # there are more than threshcount matches at the single best time skew.
        # Note: now we allow multiple matches per ID, this may need to grow
        # so it can grow inside the loop.
        results = np.zeros((len(ids), 7), np.int32)
        if not hits.size:
            # No hits found, return empty results
            return results
        # Sort hits into time_in_original order - needed for _calc_time_range
        sorted_hits = hits[hits[:, 3].argsort()]
        allids = sorted_hits[:, 0].astype(int)
        alltimes = sorted_hits[:, 1].astype(int)
        window = int(self.window)
        # Make sure every value in alltimes is >=0 for bincount
        mintime = np.amin(alltimes)
        alltimes -= mintime
        nresults = 0
        min_time = 0
        max_time = 0
        for urank, (track_id, rawcount) in enumerate(zip(ids, rawcounts, strict=False)):
            # Make sure id is an int64 before shifting it up.
            track_id = int(track_id)
            # Select the subrange of bincounts corresponding to this id
            bincounts = np.bincount(alltimes[allids == track_id])
            still_looking = True
            # Only consider legit local maxima in bincounts.
            filtered_bincounts = keep_local_maxes(bincounts)
            found_this_id = 0
            while still_looking:
                mode = int(np.argmax(filtered_bincounts))
                if filtered_bincounts[mode] <= self.threshcount:
                    # Too few - skip to the next id
                    still_looking = False
                    continue
                start = max(0, mode - window)
                stop = mode + window + 1
                count = np.sum(bincounts[start:stop])
                if self.find_time_range:
                    min_time, max_time = self._calculate_time_ranges(
                            sorted_hits, track_id, mode + mintime)
                results[nresults, :] = [track_id, count, mode + mintime, rawcount,
                                        urank, min_time, max_time]
                nresults += 1
                if nresults >= results.shape[0]:
                    results = np.vstack([results, np.zeros(results.shape,
                                                           np.int32)])
                # Clear this hit to find next largest.
                filtered_bincounts[max(0, mode - self.window):
                                   (mode + self.window + 1)] = 0
                found_this_id += 1
                if found_this_id > self.max_alignments_per_id:
                    still_looking = False
        return results[:nresults, :]

    def match_hashes(
        self,
        ht: hash_table.HashTable,
        hashes: np.ndarray,
        hashesfor: int | None = None,
    ) -> np.ndarray | tuple[np.ndarray, np.ndarray]:
        """ Match audio against fingerprint hash table.
            Return top N matches as (id, filteredmatches, timoffs, rawmatches,
            origrank, mintime, maxtime)
            If hashesfor specified, return the actual matching hashes for that
            hit (0=top hit).
        """
        # find the implicated id, time pairs from hash table
        # log("nhashes=%d" % np.shape(hashes)[0])
        hits = ht.get_hits(hashes)

        bestids, rawcounts = self._best_count_ids(hits, ht)

        # log("len(rawcounts)=%d max(rawcounts)=%d" %
        #    (len(rawcounts), max(rawcounts)))
        if not self.exact_count:
            results = self._approx_match_counts(hits, bestids, rawcounts)
        else:
            results = self._exact_match_counts(hits, bestids, rawcounts,
                                               hashesfor)
        # Sort results by filtered count, descending
        results = results[(-results[:, 1]).argsort(),]
        if hashesfor is None:
            return results
        track_id = results[hashesfor, 0]
        mode = results[hashesfor, 2]
        hashesforhashes = self._unique_match_hashes(track_id, hits, mode)
        return results, hashesforhashes

    def match_file(
        self,
        analyzer_obj: analyzer.Analyzer,
        ht: hash_table.HashTable,
        filename: str,
        number: int | None = None,
    ) -> tuple[np.ndarray, float, int]:
        """ Read in an audio file, calculate its landmarks, query against
            hash table.  Return top N matches as (id, filterdmatchcount,
            timeoffs, rawmatchcount), also length of input file in sec,
            and count of raw query hashes extracted
        """
        q_hashes = analyzer_obj.wavfile2hashes(filename)
        # Fake durations as largest hash time
        if len(q_hashes) == 0:
            durd = 0.0
        else:
            durd = analyzer_obj.n_hop * q_hashes[-1][0] / analyzer_obj.target_sr

        numberstring = "#%d" % number if number is not None else ""
        logger.debug("%s Analyzed %s %s of %.3f s to %d hashes",
                     time.ctime(), numberstring, filename, durd, len(q_hashes))

        # Run query
        rslts = cast(np.ndarray, self.match_hashes(ht, q_hashes))
        # Post filtering
        if self.sort_by_time:
            rslts = rslts[(-rslts[:, 2]).argsort(), :]
        return rslts[:self.max_returns, :], durd, len(q_hashes)

    def file_match_to_msgs(
        self,
        analyzer_obj: analyzer.Analyzer,
        ht: hash_table.HashTable,
        qry: str,
        number: int | None = None,
    ) -> list[str]:
        """ Perform a match on a single input file, return list
            of message strings """
        rslts, dur, nhash = self.match_file(analyzer_obj, ht, qry, number)
        t_hop = analyzer_obj.n_hop / analyzer_obj.target_sr
        show_verbose = logger.isEnabledFor(logging.DEBUG)
        qrymsg = f"{qry} {dur:.1f} sec {nhash} raw hashes" if show_verbose else qry
        msgrslt = []
        if len(rslts) == 0:
            # No matches returned at all
            nhashaligned = 0
            msgrslt.append(f"NOMATCH: {qrymsg}")
        else:
            for (tophitid, nhashaligned, aligntime, nhashraw, rank,
                 min_time, max_time) in rslts:
                # figure the number of raw and aligned matches for top hit
                if show_verbose:
                    if self.find_time_range:
                        msg = (
                            f"Matched {(max_time - min_time) * t_hop:6.1f} s "
                            f"starting at {min_time * t_hop:6.1f} s in {qry}"
                            f" to time {(min_time + aligntime) * t_hop:6.1f} s "
                            f"in {str(ht.names[tophitid])}"
                        )
                    else:
                        msg = f"Matched {qrymsg} as {str(ht.names[tophitid])} at {aligntime * t_hop:6.1f} s"
                    msg += f" with {nhashaligned:5d} of {nhashraw:5d} common hashes at rank {rank:2d}"
                    msgrslt.append(msg)
                else:
                    msgrslt.append(f"{qrymsg}\t{ht.names[tophitid][:]}")
                if self.illustrate:
                    self.illustrate_match(analyzer_obj, ht, qry)
        return msgrslt

    def illustrate_match(
        self,
        analyzer_obj: analyzer.Analyzer,
        ht: hash_table.HashTable,
        filename: str,
    ) -> np.ndarray:
        """ Show the query fingerprints and the matching ones
            plotted over a spectrogram """
        # Make the spectrogram
        # d, sr = librosa.load(filename, sr=analyzer.target_sr)
        d, sr = audio.audio_read(
            filename,
            sr=analyzer_obj.target_sr,
            channels=1,
            ffmpeg_path=analyzer_obj.ffmpeg_path,
        )
        sgram = np.abs(stft.stft(d, n_fft=analyzer_obj.n_fft,
                                 hop_length=analyzer_obj.n_hop,
                                 window=np.hanning(analyzer_obj.n_fft + 2)[1:-1]))
        sgram = 20.0 * np.log10(np.maximum(sgram, np.max(sgram) / 1e6))
        sgram = sgram - np.mean(sgram)
        # High-pass filter onset emphasis
        # [:-1,] discards top bin (nyquist) of sgram so bins fit in 8 bits
        # spectrogram enhancement
        if self.illustrate_hpf:
            HPF_POLE = 0.98
            sgram = np.array([scipy.signal.lfilter([1, -1],
                                                   [1, -HPF_POLE], s_row)
                              for s_row in sgram])[:-1, ]
        sgram = sgram - np.max(sgram)
        try:
            import librosa.display  # type: ignore[import-not-found]
            import matplotlib.pyplot as plt  # type: ignore[import-not-found]
        except Exception as exc:
            raise RuntimeError(
                "Plotting requires optional dependencies; install with "
                "`pip install audfprint2[plot]`."
            ) from exc
        librosa.display.specshow(sgram, sr=sr, hop_length=analyzer_obj.n_hop,
                                 y_axis='linear', x_axis='time',
                                 cmap='gray_r', vmin=-80.0, vmax=0)
        # Do the match?
        q_hashes = analyzer_obj.wavfile2hashes(filename)
        # Run query, get back the hashes for match zero
        results, matchhashes = self.match_hashes(ht, q_hashes, hashesfor=0)
        if self.sort_by_time:
            results = sorted(results, key=lambda x: -x[2])
        # Convert the hashes to landmarks
        q_hashes_list = [(int(time_), int(hash_)) for time_, hash_ in q_hashes]
        match_hashes_list = [(int(time_), int(hash_)) for time_, hash_ in matchhashes]
        lms = analyzer.hashes2landmarks(q_hashes_list)
        mlms = analyzer.hashes2landmarks(match_hashes_list)
        # Overplot on the spectrogram
        time_scale = analyzer_obj.n_hop / float(sr)
        freq_scale = float(sr)/analyzer_obj.n_fft
        plt.plot(time_scale * np.array([[x[0], x[0] + x[3]] for x in lms]).T,
                 freq_scale * np.array([[x[1], x[2]] for x in lms]).T,
                 '.-g')
        plt.plot(time_scale * np.array([[x[0], x[0] + x[3]] for x in mlms]).T,
                 freq_scale * np.array([[x[1], x[2]] for x in mlms]).T,
                 '.-r')
        # Add title
        plt.title(
            (
                f"{filename} : Matched as {ht.names[results[0][0]]}"
                + " with %d of %d hashes" % (len(matchhashes), len(q_hashes))
            )
        )
        # Display
        plt.show()
        # Return
        return results


def localtest() -> None:
    """Function to provide quick test"""
    pat = '/Users/dpwe/projects/shazam/Nine_Lives/*mp3'
    qry = 'query.mp3'
    hash_tab = analyzer.glob2hashtable(pat)
    matcher = Matcher()
    g2h = cast(analyzer.Analyzer, analyzer.g2h_analyzer)
    rslts, dur, nhash = matcher.match_file(g2h, hash_tab, qry)
    t_hop = 0.02322
    logger.debug(
        f"Matched {qry} ({dur} s, {nhash} hashes) as "
        f"{hash_tab.names[rslts[0][0]]} at {t_hop * float(rslts[0][2])} "
        f"with {rslts[0][1]} of {rslts[0][3]} hashes"
    )


# Run the main function if called from the command line
if __name__ == "__main__":
    localtest()
